#include "CellMLToRDFTree.h"
#include <nsComponentManagerUtils.h>
#include <nsServiceManagerUtils.h>
#include <nsStringAPI.h>
#include <prprf.h>

namespace iface
{
  namespace XPCOM
  {
    class IObject;
  };
};

#include <IWrappedPCM.h>

bool CellMLToRDFTree::mSetupStatic = false;
nsIRDFResource* CellMLToRDFTree::mRootResource = nsnull;
nsIRDFResource* CellMLToRDFTree::mIvRootResource = nsnull;
nsIRDFResource* CellMLToRDFTree::mTypeResource = nsnull;
nsIRDFResource* CellMLToRDFTree::mUnitsResource = nsnull;
nsIRDFResource* CellMLToRDFTree::mPropResource = nsnull;
nsIRDFResource* CellMLToRDFTree::mValueResource = nsnull;
nsIRDFResource* CellMLToRDFTree::mValuePropResource = nsnull;
nsIRDFResource* CellMLToRDFTree::mOpenResource = nsnull;
nsIRDFResource* CellMLToRDFTree::mPlaceholderResource = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mTrueLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mNumberLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mModelLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mComponentsLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mGroupLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mGroupsLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mConnectionsLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mConnectionLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mUnitsLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mImportsLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mVariablesLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mRolesLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mPubIntLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mPrivIntLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mBaseLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mDerivedLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mPrefixLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mExponentLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mOffsetLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mMultiplierLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mUnitLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mReversibleLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mIrreversibleLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mReactionLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mVariableRefLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mReactantLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mProductLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mRateLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mCatalystLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mActivatorLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mInhibitorLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mModifierLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mDirectionLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mForwardLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mReverseLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mBothLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mStoichiometryLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mDeltaVariableLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mRelationshipRefLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mNameLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mNamespaceLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mComponentRefLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mComponent1Literal = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mComponent2Literal = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mVariable1Literal = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mVariable2Literal = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mVariableMappingLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mMapVariablesLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mMathLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mEquationLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mInterfaceIdLiteral[3];
nsCOMPtr<nsIRDFService> CellMLToRDFTree::mRDFService;
nsCOMPtr<nsIRDFContainerUtils> CellMLToRDFTree::mContUtils;
nsCOMPtr<nsIRDFDataSource> CellMLToRDFTree::mPersistSource;

CellMLToRDFTree::CellMLToRDFTree()
{
  mInvertOpenSet.Init();
  mHash.Init();
  mResourceHash.Init();
}

CellMLToRDFTree::~CellMLToRDFTree()
{
}

NS_IMPL_ISUPPORTS1(CellMLToRDFTree, ICellMLToRDFTree);


void
CellMLToRDFTree::EnsureLiterals()
{
  if (mSetupStatic)
    return;
  mSetupStatic = PR_TRUE;
  mRDFService = do_GetService("@mozilla.org/rdf/rdf-service;1");
  mContUtils = do_GetService("@mozilla.org/rdf/container-utils;1");
#define RESOURCE(v,url) \
  mRDFService->GetResource(NS_LITERAL_CSTRING(url), &m##v##Resource)

  RESOURCE(Root, "http://www.cellml.org/tools/pcenv/model_tree_view");
  RESOURCE(IvRoot, "http://www.cellml.org/tools/pcenv/model_tree_view#ivonly");
  RESOURCE(Type, "http://www.cellml.org/tools/pcenv/model_tree_view#type");
  RESOURCE(Units, "http://www.cellml.org/tools/pcenv/model_tree_view#units");
  RESOURCE(Prop, "http://www.cellml.org/tools/pcenv/model_tree_view#prop");
  RESOURCE(Value, "http://www.cellml.org/tools/pcenv/model_tree_view#value");
  RESOURCE(ValueProp, "http://www.cellml.org/tools/pcenv/model_tree_view#valueprop");
  RESOURCE(Open, "http://home.netscape.com/NC-rdf#open");
  RESOURCE(Placeholder, "http://www.cellml.org/tools/pcenv/placeholder");

#undef RESOURCE
#define LITERAL(v,val) \
  mRDFService->GetLiteral(NS_LITERAL_STRING(val).get(), &m##v##Literal)

  LITERAL(True, "true");
  LITERAL(Number, "number");
  LITERAL(Model, "Model");
  LITERAL(Components, "Components");
  LITERAL(Group, "Group");
  LITERAL(Groups, "Groups");
  LITERAL(Connections, "Connections");
  LITERAL(Connection, "Connection");
  LITERAL(Units, "Units");
  LITERAL(Imports, "Imports");
  LITERAL(Variables, "Variables");
  LITERAL(Roles, "Roles");
  LITERAL(PubInt, "Public Interface");
  LITERAL(PrivInt, "Private Interface");
  LITERAL(Base, "Base");
  LITERAL(Derived, "Derived");
  LITERAL(Prefix, "Prefix");
  LITERAL(Exponent, "Exponent");
  LITERAL(Offset, "Offset");
  LITERAL(Multiplier, "Multiplier");
  LITERAL(Unit, "Unit");
  LITERAL(Reversible, "Reversible");
  LITERAL(Irreversible, "Irreversible");
  LITERAL(Reaction, "Reaction");
  LITERAL(VariableRef, "Variable Reference");
  LITERAL(Reactant, "Reactant");
  LITERAL(Product, "Product");
  LITERAL(Rate, "Rate");
  LITERAL(Catalyst, "Catalyst");
  LITERAL(Activator, "Activator");
  LITERAL(Inhibitor, "Inhibitor");
  LITERAL(Modifier, "Modifier");
  LITERAL(Direction, "Direction");
  LITERAL(Forward, "Forward");
  LITERAL(Reverse, "Reverse");
  LITERAL(Both, "Both");
  LITERAL(Stoichiometry, "Stoichiometry");
  LITERAL(DeltaVariable, "Delta Variable");
  LITERAL(RelationshipRef, "RelationshipRef");
  LITERAL(Name, "Name");
  LITERAL(Namespace, "Namespace");
  LITERAL(ComponentRef, "ComponentRef");
  LITERAL(Component1, "Component 1");
  LITERAL(Component2, "Component 2");
  LITERAL(Variable1, "Variable 1");
  LITERAL(Variable2, "Variable 2");
  LITERAL(VariableMapping, "Variable Mapping");
  LITERAL(MapVariables, "MapVariables");
  LITERAL(Math, "Mathematics");
  LITERAL(Equation, "Equation");

#undef LITERAL

  mRDFService->GetLiteral(NS_LITERAL_STRING("In").get(), mInterfaceIdLiteral);
  mRDFService->GetLiteral(NS_LITERAL_STRING("Out").get(),
                          mInterfaceIdLiteral + 1);
  mRDFService->GetLiteral(NS_LITERAL_STRING("None").get(),
                          mInterfaceIdLiteral + 2);

  mRDFService->GetDataSource("rdf:local-store", getter_AddRefs(mPersistSource));
}

void
CellMLToRDFTree::ExpandResourceInTree(nsIRDFResource* aResource)
{
  mPersistSource->Assert(aResource, mOpenResource, mTrueLiteral, true);
}

NS_IMETHODIMP
CellMLToRDFTree::Initialise
(
 cellml_apiIModel* aModel
)
{
  NS_ENSURE_TRUE(aModel,NS_ERROR_INVALID_ARG);
  EnsureLiterals();

  mHash.Clear();
  mResourceHash.Clear();

  // Make a new datasource...
  mDatasource = do_CreateInstance
    ("@mozilla.org/rdf/datasource;1?name=in-memory-datasource");
  if (mDatasource == nsnull)
    return NS_ERROR_OUT_OF_MEMORY;

  // Make top-level sequences...
  nsCOMPtr<nsIRDFContainer> rootseq, ivseq;
  mContUtils->MakeSeq(mDatasource, mRootResource, getter_AddRefs(rootseq));
  mContUtils->MakeSeq(mDatasource, mIvRootResource, getter_AddRefs(ivseq));

  PopulateModelIntoSequences(aModel, rootseq);

  nsCOMPtr<cellml_apiICellMLComponentSet> comps;
  aModel->GetLocalComponents(getter_AddRefs(comps));
  nsCOMPtr<cellml_apiICellMLComponentIterator> compi;
  comps->IterateComponents(getter_AddRefs(compi));
  while (1)
  {
    nsCOMPtr<cellml_apiICellMLComponent> comp;
    compi->NextComponent(getter_AddRefs(comp));
    if (comp == nsnull)
      break;
    nsCOMPtr<cellml_apiICellMLVariableSet> vars;
    comp->GetVariables(getter_AddRefs(vars));
    nsCOMPtr<cellml_apiICellMLVariableIterator> vari;
    vars->IterateVariables(getter_AddRefs(vari));
    PopulateVariablesIntoSequences(vari, ivseq, PR_TRUE);
  }

  return NS_OK;
}

CellMLToRDFTree::ResourceAssociation::ResourceAssociation
(
 const char* aType,
 nsISupports* aElement,
 CellMLToRDFTree* aCellMLToRDFTree,
 nsIRDFContainer* aSeq,
 rdf_populate_t aPopulate,
 PRBool aDefaultOpen
)
  : mType(aType), mElement(aElement), mSeq(aSeq),
    mCellMLToRDFTree(aCellMLToRDFTree), mPopulate(aPopulate),
    mDefaultOpen(aDefaultOpen)
{
}

void MakeAssociationKey(const char* aType,
                        nsISupports* aElement,
                        nsACString& aStr)
{
  nsCString tmp;
  nsCOMPtr<IWrappedPCM> cw(do_QueryInterface(aElement));
  cw->GetObjid(tmp);
  aStr.Assign(tmp);
  aStr.AppendLiteral(":");
  aStr.Append(aType);
}

already_AddRefed<nsIRDFResource>
CellMLToRDFTree::GetAssociatedResource
(
 const char* aType,
 nsISupports* aElement,
 rdf_populate_t aPopulate,
 PRBool aDefaultOpen
)
{
  nsCString assocKey;
  MakeAssociationKey(aType, aElement, assocKey);

  nsCOMPtr<nsIRDFResource> blank;
  nsIRDFResource* ret;
  mResourceHash.Get(assocKey, getter_AddRefs(blank));
  if (blank != nsnull)
  {
    ret = blank;
    NS_ADDREF(ret);
    return ret;
  }

  mRDFService->GetAnonymousResource(getter_AddRefs(blank));

  nsCOMPtr<nsIRDFContainer> seq;
  if (aPopulate != nsnull)
    mContUtils->MakeSeq(mDatasource, blank, getter_AddRefs(seq));

  int val = 0;
  mInvertOpenSet.Get(assocKey, &val);
  PRBool isOpen = aDefaultOpen ? (val == 0) : (val == 1);
  if (isOpen)
  {
    aPopulate(this, aElement, seq);
    ExpandResourceInTree(blank);
    aPopulate = PR_FALSE;
  }
  else if (aPopulate != nsnull)
  {
    seq->AppendElement(mPlaceholderResource);
  }

  ResourceAssociation* ra =
    new ResourceAssociation(aType, aElement, this, seq, aPopulate, aDefaultOpen);

  ret = blank;
  mHash.Put(ret, ra);
  mResourceHash.Put(assocKey, ret);

  NS_ADDREF(ret);
  return ret;
}

static void
populateModelComponents(CellMLToRDFTree* aTree,
                        nsISupports* aElement,
                        nsIRDFContainer* aSeq)
{
  nsCOMPtr<cellml_apiIModel> mod(do_QueryInterface(aElement));
  nsCOMPtr<cellml_apiICellMLComponentSet> comps;
  nsCOMPtr<cellml_apiICellMLComponentIterator> compi;
  mod->GetLocalComponents(getter_AddRefs(comps));
  comps->IterateComponents(getter_AddRefs(compi));
  aTree->PopulateComponentsIntoSequences(compi, aSeq);
}

static void
populateModelGroups(CellMLToRDFTree* aTree,
                    nsISupports* aElement,
                    nsIRDFContainer* aSeq)
{
  nsCOMPtr<cellml_apiIModel> mod(do_QueryInterface(aElement));
  nsCOMPtr<cellml_apiIGroupSet> groupss;
  nsCOMPtr<cellml_apiIGroupIterator> groupsi;
  mod->GetGroups(getter_AddRefs(groupss));
  groupss->IterateGroups(getter_AddRefs(groupsi));
  aTree->PopulateGroupsIntoSequences(groupsi, aSeq);
}

static void
populateModelConnections(CellMLToRDFTree* aTree,
                         nsISupports* aElement,
                         nsIRDFContainer* aSeq)
{
  nsCOMPtr<cellml_apiIModel> mod(do_QueryInterface(aElement));
  nsCOMPtr<cellml_apiIConnectionSet> conns;
  nsCOMPtr<cellml_apiIConnectionIterator> conni;
  mod->GetConnections(getter_AddRefs(conns));
  conns->IterateConnections(getter_AddRefs(conni));
  aTree->PopulateConnectionsIntoSequences(conni, aSeq);
}

static void
populateModelUnits(CellMLToRDFTree* aTree,
                   nsISupports* aElement,
                   nsIRDFContainer* aSeq)
{
  nsCOMPtr<cellml_apiIModel> mod(do_QueryInterface(aElement));
  nsCOMPtr<cellml_apiIUnitsSet> unitss;
  nsCOMPtr<cellml_apiIUnitsIterator> unitsi;
  mod->GetLocalUnits(getter_AddRefs(unitss));
  unitss->IterateUnits(getter_AddRefs(unitsi));
  aTree->PopulateUnitsIntoSequences(unitsi, aSeq);
}

static void
populateModelImports(CellMLToRDFTree* aTree,
                     nsISupports* aElement,
                     nsIRDFContainer* aSeq)
{
  nsCOMPtr<cellml_apiIModel> mod(do_QueryInterface(aElement));
  nsCOMPtr<cellml_apiICellMLImportSet> importss;
  nsCOMPtr<cellml_apiICellMLImportIterator> importsi;
  mod->GetImports(getter_AddRefs(importss));
  importss->IterateImports(getter_AddRefs(importsi));
  aTree->PopulateImportsIntoSequences(importsi, aSeq);
}

void
CellMLToRDFTree::PopulateModelIntoSequences
(
 cellml_apiIModel* aModel,
 nsIRDFContainer* aRootseq
)
{
  // Create the top-level entries...
#define TOPLEVEL(x,X) \
  nsCOMPtr<nsIRDFResource> x(\
    GetAssociatedResource("model_" #x, aModel,\
                          &populateModel##X, PR_TRUE)); \
  nsCOMPtr<nsIRDFContainer> x##Seq; \
  mDatasource->Assert(x, mTypeResource, m##X##Literal, true); \
  mDatasource->Assert(x, mPropResource, m##X##Literal, true); \
  aRootseq->AppendElement(x);

  TOPLEVEL(components,Components);
  TOPLEVEL(groups,Groups);
  TOPLEVEL(connections,Connections);
  TOPLEVEL(units,Units);
  TOPLEVEL(imports,Imports);
#undef TOPLEVEL
}

static void
populateComponent(CellMLToRDFTree* aTree,
                  nsISupports* aElement,
                  nsIRDFContainer* aSeq)
{
  nsCOMPtr<cellml_apiICellMLComponent> comp(do_QueryInterface(aElement));
  nsCOMPtr<cellml_apiICellMLVariableSet> vars;
  comp->GetVariables(getter_AddRefs(vars));
  nsCOMPtr<cellml_apiICellMLVariableIterator> vari;
  vars->IterateVariables(getter_AddRefs(vari));
  aTree->PopulateVariablesIntoSequences(vari, aSeq);

  nsCOMPtr<cellml_apiIUnitsSet> units;
  comp->GetUnits(getter_AddRefs(units));
  nsCOMPtr<cellml_apiIUnitsIterator> unitsi;
  units->IterateUnits(getter_AddRefs(unitsi));
  aTree->PopulateUnitsIntoSequences(unitsi, aSeq);
  
  nsCOMPtr<cellml_apiIReactionSet> reactions;
  comp->GetReactions(getter_AddRefs(reactions));
  nsCOMPtr<cellml_apiIReactionIterator> reacti;
  reactions->IterateReactions(getter_AddRefs(reacti));
  aTree->PopulateReactionsIntoSequences(reacti, aSeq);

  nsCOMPtr<cellml_apiIMathContainer> mc(do_QueryInterface(aElement));
  nsCOMPtr<cellml_apiIMathList> ml;
  mc->GetMath(getter_AddRefs(ml));
  nsCOMPtr<cellml_apiIMathMLElementIterator> mmi;
  ml->Iterate(getter_AddRefs(mmi));
  aTree->PopulateMathsIntoSequences(mmi, aSeq);
}

void
CellMLToRDFTree::PopulateComponentsIntoSequences
(
 cellml_apiICellMLComponentIterator* aCompIt,
 nsIRDFContainer* aCompSeq
)
{
  nsCOMPtr<cellml_apiICellMLComponent> comp;
  while (PR_TRUE)
  {
    aCompIt->NextComponent(getter_AddRefs(comp));
    if (comp == nsnull)
      break;
    nsCOMPtr<nsIRDFResource> resource
      (GetAssociatedResource("component", comp, &populateComponent));

    nsCOMPtr<nsIRDFLiteral> nameLit;
    nsString compNameStr;
    comp->GetName(compNameStr);
    mRDFService->GetLiteral(compNameStr.get(), getter_AddRefs(nameLit));

    mDatasource->Assert(resource, mTypeResource, nameLit, true);
    mDatasource->Assert(resource, mPropResource, mComponentsLiteral, true);

    aCompSeq->AppendElement(resource);
  }
}

static void
populateGroup(CellMLToRDFTree* aTree,
              nsISupports* aElement,
              nsIRDFContainer* aSeq)
{
  nsCOMPtr<cellml_apiIGroup> group(do_QueryInterface(aElement));
  nsCOMPtr<cellml_apiIRelationshipRefSet> rrs;
  group->GetRelationshipRefs(getter_AddRefs(rrs));
  nsCOMPtr<cellml_apiIRelationshipRefIterator> rri;
  rrs->IterateRelationshipRefs(getter_AddRefs(rri));
  aTree->PopulateRelationshipRefsIntoSequences(rri, aSeq);
    
  nsCOMPtr<cellml_apiIComponentRefSet> crs;
  group->GetComponentRefs(getter_AddRefs(crs));
  nsCOMPtr<cellml_apiIComponentRefIterator> cri;
  crs->IterateComponentRefs(getter_AddRefs(cri));
  aTree->PopulateComponentRefsIntoSequences(cri, aSeq);
  
}

void
CellMLToRDFTree::PopulateGroupsIntoSequences
(
 cellml_apiIGroupIterator* aGroupIt,
 nsIRDFContainer* aGroupSeq
)
{
  nsCOMPtr<cellml_apiIGroup> group;
  while (PR_TRUE)
  {
    aGroupIt->NextGroup(getter_AddRefs(group));

    if (group == nsnull)
      break;

    nsCOMPtr<nsIRDFResource> resource
      (GetAssociatedResource("group", group, populateGroup));

    mDatasource->Assert(resource, mTypeResource, mGroupLiteral, PR_TRUE);
    mDatasource->Assert(resource, mPropResource, mGroupsLiteral, PR_TRUE);
    
    aGroupSeq->AppendElement(resource);
  }
}

static void
populateConnection(CellMLToRDFTree* aTree,
                   nsISupports* aElement,
                   nsIRDFContainer* aSeq)
{
  nsCOMPtr<cellml_apiIConnection> conn(do_QueryInterface(aElement));
  nsCOMPtr<cellml_apiIMapComponents> mapcom;
  conn->GetComponentMapping(getter_AddRefs(mapcom));

  nsCOMPtr<nsIRDFResource> mapCompRes
    (aTree->GetAssociatedResource("connection_comp1", conn, nsnull));
  aTree->mDatasource->Assert(mapCompRes, aTree->mTypeResource,
                             aTree->mComponent1Literal, PR_TRUE);
  nsCOMPtr<nsIRDFLiteral> lit;
  nsString str;
  mapcom->GetFirstComponentName(str);
  aTree->mRDFService->GetLiteral(str.get(), getter_AddRefs(lit));
  aTree->mDatasource->Assert(mapCompRes, aTree->mValueResource, lit, PR_TRUE);
  aSeq->AppendElement(mapCompRes);

  mapCompRes = aTree->GetAssociatedResource("connection_comp2", conn, nsnull);
  aTree->mDatasource->Assert(mapCompRes, aTree->mTypeResource,
                             aTree->mComponent2Literal, PR_TRUE);
  mapcom->GetSecondComponentName(str);
  aTree->mRDFService->GetLiteral(str.get(), getter_AddRefs(lit));
  aTree->mDatasource->Assert(mapCompRes, aTree->mValueResource, lit, PR_TRUE);
  aSeq->AppendElement(mapCompRes);
  
  nsCOMPtr<cellml_apiIMapVariablesSet> mvs;
  conn->GetVariableMappings(getter_AddRefs(mvs));
  nsCOMPtr<cellml_apiIMapVariablesIterator> mvi;
  mvs->IterateMapVariables(getter_AddRefs(mvi));
    
  aTree->PopulateMapVariablesIntoSequences(mvi, aSeq);
}

void
CellMLToRDFTree::PopulateConnectionsIntoSequences
(
 cellml_apiIConnectionIterator* aConnIt,
 nsIRDFContainer* aConnSeq
)
{
  nsCOMPtr<cellml_apiIConnection> conn;
  while (PR_TRUE)
  {
    aConnIt->NextConnection(getter_AddRefs(conn));
    if (conn == nsnull)
      break;
    nsCOMPtr<nsIRDFResource> resource
      (GetAssociatedResource("connection", conn, &populateConnection));

    mDatasource->Assert(resource, mTypeResource, mConnectionLiteral, PR_TRUE);
    mDatasource->Assert(resource, mPropResource, mConnectionsLiteral, PR_TRUE);

    aConnSeq->AppendElement(resource);
  }
}

static void
populateUnits(CellMLToRDFTree* aTree,
              nsISupports* aElement,
              nsIRDFContainer* aSeq)
{
  nsCOMPtr<cellml_apiIUnits> units(do_QueryInterface(aElement));
  nsCOMPtr<cellml_apiIUnitSet> uns;
  units->GetUnitCollection(getter_AddRefs(uns));
  nsCOMPtr<cellml_apiIUnitIterator> uni;
  uns->IterateUnits(getter_AddRefs(uni));
  aTree->PopulateUnitIntoSequences(uni, aSeq);
}

void
CellMLToRDFTree::PopulateUnitsIntoSequences
(
 cellml_apiIUnitsIterator* aUnitsIt,
 nsIRDFContainer* aUnitsSeq
)
{
  nsCOMPtr<cellml_apiIUnits> units;
  while (PR_TRUE)
  {
    aUnitsIt->NextUnits(getter_AddRefs(units));
    if (units == nsnull)
      break;
    nsCOMPtr<nsIRDFResource> resource
      (GetAssociatedResource("units", units, &populateUnits));

    nsString str;
    units->GetName(str);
    nsCOMPtr<nsIRDFLiteral> lit;
    mRDFService->GetLiteral(str.get(), getter_AddRefs(lit));
    mDatasource->Assert(resource, mTypeResource, lit, PR_TRUE);
    mDatasource->Assert(resource, mPropResource, mUnitsLiteral, PR_TRUE);
    PRBool isBase;
    units->GetIsBaseUnits(&isBase);
    mDatasource->Assert(resource, mValueResource, isBase ? mBaseLiteral :
                        mDerivedLiteral, PR_TRUE);

    aUnitsSeq->AppendElement(resource);
  }
}

static void
populateImportComponents(CellMLToRDFTree* aTree,
                         nsISupports* aElement,
                         nsIRDFContainer* aSeq)
{
  nsCOMPtr<cellml_apiICellMLImport> import(do_QueryInterface(aElement));
  nsCOMPtr<cellml_apiIImportComponentSet> ics;
  import->GetComponents(getter_AddRefs(ics));
  nsCOMPtr<cellml_apiIImportComponentIterator> ici;
  ics->IterateImportComponents(getter_AddRefs(ici));
  aTree->PopulateImportComponentsIntoSequences(ici, aSeq);
}

static void
populateImportUnits(CellMLToRDFTree* aTree,
                    nsISupports* aElement,
                    nsIRDFContainer* aSeq)
{
  nsCOMPtr<cellml_apiICellMLImport> import(do_QueryInterface(aElement));
  nsCOMPtr<cellml_apiIImportUnitsSet> ius;
  import->GetUnits(getter_AddRefs(ius));
  nsCOMPtr<cellml_apiIImportUnitsIterator> iui;
  ius->IterateImportUnits(getter_AddRefs(iui));
  aTree->PopulateImportUnitsIntoSequences(iui, aSeq);
}

static void
populateImportModel(CellMLToRDFTree* aTree,
                    nsISupports* aElement,
                    nsIRDFContainer* aSeq)
{
  nsCOMPtr<cellml_apiIModel> mod(do_QueryInterface(aElement));
  aTree->PopulateModelIntoSequences(mod, aSeq);
}

static void
populateImport(CellMLToRDFTree* aTree,
               nsISupports* aElement,
               nsIRDFContainer* aSeq)
{
  nsCOMPtr<cellml_apiICellMLImport> import(do_QueryInterface(aElement));
  nsCOMPtr<nsIRDFResource> icresource
    (aTree->GetAssociatedResource("import_components", import,
                                  &populateImportComponents));
  aTree->mDatasource->Assert(icresource, aTree->mTypeResource,
                             aTree->mComponentsLiteral, PR_TRUE);
  aTree->mDatasource->Assert(icresource, aTree->mPropResource,
                             aTree->mComponentsLiteral, PR_TRUE);
  aSeq->AppendElement(icresource);

  nsCOMPtr<nsIRDFResource> iuresource
    (aTree->GetAssociatedResource("import_units", import,
                                  &populateImportUnits));
  aTree->mDatasource->Assert(iuresource, aTree->mTypeResource,
                             aTree->mUnitsLiteral, PR_TRUE);
  aTree->mDatasource->Assert(iuresource, aTree->mPropResource,
                             aTree->mUnitsLiteral, PR_TRUE);
  aSeq->AppendElement(iuresource);

  nsCOMPtr<cellml_apiIModel> mod;
  import->GetImportedModel(getter_AddRefs(mod));
  if (mod != nsnull)
  {
    nsCOMPtr<nsIRDFResource> modresource
      (aTree->GetAssociatedResource("import_model", mod,
                                    &populateImportModel));
    aTree->mDatasource->Assert(modresource, aTree->mTypeResource,
                               aTree->mModelLiteral, PR_TRUE);
    aTree->mDatasource->Assert(modresource, aTree->mPropResource,
                               aTree->mModelLiteral, PR_TRUE);
    aSeq->AppendElement(modresource);
  }
}

void
CellMLToRDFTree::PopulateImportsIntoSequences
(
 cellml_apiICellMLImportIterator* aImpIt,
 nsIRDFContainer* aImpSeq
)
{
  nsCOMPtr<cellml_apiICellMLImport> import;
  while (PR_TRUE)
  {
    aImpIt->NextImport(getter_AddRefs(import));
    if (import == nsnull)
      break;
    nsCOMPtr<nsIRDFResource> resource
      (GetAssociatedResource("import", import, &populateImport));

    nsString str;
    nsCOMPtr<cellml_apiIURI> xlink;
    import->GetXlinkHref(getter_AddRefs(xlink));
    xlink->GetAsText(str);
    nsCOMPtr<nsIRDFLiteral> lit;
    mRDFService->GetLiteral(str.get(), getter_AddRefs(lit));
    mDatasource->Assert(resource, mTypeResource, lit, PR_TRUE);
    mDatasource->Assert(resource, mPropResource, mImportsLiteral, PR_TRUE);

    aImpSeq->AppendElement(resource);
  }
}

static void
populateVariable(CellMLToRDFTree* aTree,
                 nsISupports* aElement,
                 nsIRDFContainer* aSeq)
{
  nsCOMPtr<cellml_apiICellMLVariable> var(do_QueryInterface(aElement));
    
  nsCOMPtr<nsIRDFResource> pubIntRes
    (aTree->GetAssociatedResource("public_interface", var, nsnull));
  aTree->mDatasource->Assert(pubIntRes, aTree->mTypeResource,
                             aTree->mPubIntLiteral, PR_TRUE);
  PRUint32 pubint;
  var->GetPublicInterface(&pubint);
  aTree->mDatasource->Assert(pubIntRes, aTree->mValueResource,
                             aTree->mInterfaceIdLiteral[pubint], PR_TRUE);
  aSeq->AppendElement(pubIntRes);

  nsCOMPtr<nsIRDFResource> privIntRes
    (aTree->GetAssociatedResource("private_interface", var, nsnull));
  aTree->mDatasource->Assert(privIntRes, aTree->mTypeResource,
                             aTree->mPrivIntLiteral, PR_TRUE);
  PRUint32 privint;
  var->GetPrivateInterface(&privint);
  aTree->mDatasource->Assert(privIntRes, aTree->mValueResource,
                             aTree->mInterfaceIdLiteral[privint], PR_TRUE);
  aSeq->AppendElement(privIntRes);
}

void
CellMLToRDFTree::PopulateVariablesIntoSequences
(
 cellml_apiICellMLVariableIterator* aVarIt,
 nsIRDFContainer* aVarSeq,
 PRBool aIgnoreNonIV
)
{
  nsCOMPtr<cellml_apiICellMLVariable> var;
  while (PR_TRUE)
  {
    aVarIt->NextVariable(getter_AddRefs(var));
    if (var == nsnull)
      break;
    nsCOMPtr<nsIRDFResource> resource
      (GetAssociatedResource("variable", var, &populateVariable));
    nsString str;
    var->GetName(str);
    nsCOMPtr<nsIRDFLiteral> lit;
    mRDFService->GetLiteral(str.get(), getter_AddRefs(lit));
    mDatasource->Assert(resource, mTypeResource, lit, PR_TRUE);
    mDatasource->Assert(resource, mPropResource, mVariablesLiteral, PR_TRUE);

    var->GetUnitsName(str);
    mRDFService->GetLiteral(str.get(), getter_AddRefs(lit));
    mDatasource->Assert(resource, mUnitsResource, lit, PR_TRUE);

    nsString iv;
    nsresult ret;
    ret = var->GetInitialValue(iv);
    if (NS_SUCCEEDED(ret) && !iv.Equals(NS_LITERAL_STRING("")))
    {
      mRDFService->GetLiteral(iv.get(), getter_AddRefs(lit));
      mDatasource->Assert(resource, mValueResource, lit, PR_TRUE);
      mDatasource->Assert(resource, mValuePropResource, mNumberLiteral,
                          PR_TRUE);
      // If the variable has an initial value, put it into the IV seq too...
      // aIVseq->AppendElement(resource);
      if (aIgnoreNonIV)
        aVarSeq->AppendElement(resource);
    }
    if (!aIgnoreNonIV)
      aVarSeq->AppendElement(resource);
  }
}

static void
populateReaction(CellMLToRDFTree* aTree,
                 nsISupports* aElement,
                 nsIRDFContainer* aSeq)
{
  nsCOMPtr<cellml_apiIReaction> react(do_QueryInterface(aElement));
  nsCOMPtr<cellml_apiIVariableRefSet> vrs;
  react->GetVariableReferences(getter_AddRefs(vrs));
  nsCOMPtr<cellml_apiIVariableRefIterator> vri;
  vrs->IterateVariableRefs(getter_AddRefs(vri));
  aTree->PopulateVariableRefsIntoSequences(vri, aSeq);

}

void
CellMLToRDFTree::PopulateReactionsIntoSequences
(
 cellml_apiIReactionIterator* aReactIt,
 nsIRDFContainer* aReactSeq
)
{
  nsCOMPtr<cellml_apiIReaction> react;
  while (PR_TRUE)
  {
    aReactIt->NextReaction(getter_AddRefs(react));
    if (react == nsnull)
      break;
    nsCOMPtr<nsIRDFResource> resource
      (GetAssociatedResource("reaction", react, &populateReaction));

    PRBool isReversible;
    react->GetReversible(&isReversible);
    mDatasource->Assert(resource, mTypeResource,
                        isReversible ? mReversibleLiteral :
                        mIrreversibleLiteral, PR_TRUE);
    mDatasource->Assert(resource, mPropResource,
                        mReactionLiteral, PR_TRUE);
    aReactSeq->AppendElement(resource);
  }
}

static void
populateRelationshipRef(CellMLToRDFTree* aTree,
                        nsISupports* aElement,
                        nsIRDFContainer* aSeq)
{
  nsCOMPtr<cellml_apiIRelationshipRef> rr(do_QueryInterface(aElement));
  nsCOMPtr<nsIRDFResource> namersrc
    (aTree->GetAssociatedResource("relationshipRef_name", rr, nsnull));
  aTree->mDatasource->Assert(namersrc, aTree->mTypeResource,
                             aTree->mNameLiteral, PR_TRUE);
  nsString str;
  nsCOMPtr<nsIRDFLiteral> lit;
  rr->GetName(str);
  aTree->mRDFService->GetLiteral(str.get(), getter_AddRefs(lit));
  aTree->mDatasource->Assert(namersrc, aTree->mValueResource, lit, PR_TRUE);
  aSeq->AppendElement(namersrc);

  nsCOMPtr<nsIRDFResource> nsrsrc
    (aTree->GetAssociatedResource("relationshipRef_namespace", rr,
                                  nsnull));
  aTree->mDatasource->Assert(nsrsrc, aTree->mTypeResource,
                             aTree->mNamespaceLiteral, PR_TRUE);
  rr->GetRelationshipNamespace(str);
  aTree->mRDFService->GetLiteral(str.get(), getter_AddRefs(lit));
  aTree->mDatasource->Assert(nsrsrc, aTree->mValueResource, lit, PR_TRUE);
  aSeq->AppendElement(nsrsrc);
}

void
CellMLToRDFTree::PopulateRelationshipRefsIntoSequences
(
 cellml_apiIRelationshipRefIterator* aRRIt,
 nsIRDFContainer* aRRSeq
)
{
  nsCOMPtr<cellml_apiIRelationshipRef> rr;
  while (PR_TRUE)
  {
    aRRIt->NextRelationshipRef(getter_AddRefs(rr));
    if (rr == nsnull)
      break;
    nsCOMPtr<nsIRDFResource> resource
      (GetAssociatedResource("relationshipRef", rr, &populateRelationshipRef));

    nsString str;
    rr->GetRelationship(str);
    nsCOMPtr<nsIRDFLiteral> lit;
    mRDFService->GetLiteral(str.get(), getter_AddRefs(lit));

    mDatasource->Assert(resource, mTypeResource, lit, PR_TRUE);
    mDatasource->Assert(resource, mPropResource, mRelationshipRefLiteral,
                        PR_TRUE);
    aRRSeq->AppendElement(resource);
  }
}

static void
populateComponentRef(CellMLToRDFTree* aTree,
                     nsISupports* aElement,
                     nsIRDFContainer* aSeq)
{
  nsCOMPtr<cellml_apiIComponentRef> cr(do_QueryInterface(aElement));
  nsCOMPtr<cellml_apiIComponentRefSet> crs;
  cr->GetComponentRefs(getter_AddRefs(crs));
  nsCOMPtr<cellml_apiIComponentRefIterator> cri;
  crs->IterateComponentRefs(getter_AddRefs(cri));
  aTree->PopulateComponentRefsIntoSequences(cri, aSeq);
}

void
CellMLToRDFTree::PopulateComponentRefsIntoSequences
(
 cellml_apiIComponentRefIterator* aCRIt,
 nsIRDFContainer* aCRSeq
)
{
  nsCOMPtr<cellml_apiIComponentRef> cr;
  while (PR_TRUE)
  {
    aCRIt->NextComponentRef(getter_AddRefs(cr));
    if (cr == nsnull)
      break;
    nsCOMPtr<nsIRDFResource> resource
      (GetAssociatedResource("component_ref", cr, &populateComponentRef));
    
    nsString str;
    cr->GetComponentName(str);
    nsCOMPtr<nsIRDFLiteral> lit;
    mRDFService->GetLiteral(str.get(), getter_AddRefs(lit));
    mDatasource->Assert(resource, mTypeResource, lit, PR_TRUE);
    mDatasource->Assert(resource, mPropResource, mComponentRefLiteral,
                        PR_TRUE);
    aCRSeq->AppendElement(resource);
  }
}

static void
populateMapVariables(CellMLToRDFTree* aTree,
                     nsISupports* aElement,
                     nsIRDFContainer* aSeq)
{
  nsCOMPtr<cellml_apiIMapVariables> mv(do_QueryInterface(aElement));

  nsCOMPtr<nsIRDFResource> v1rsrc
    (aTree->GetAssociatedResource("map_variable_var1", mv, nsnull));
  aTree->mDatasource->Assert(v1rsrc, aTree->mTypeResource,
                             aTree->mVariable1Literal, PR_TRUE);
  nsString str;
  mv->GetFirstVariableName(str);
  nsCOMPtr<nsIRDFLiteral> lit;
  aTree->mRDFService->GetLiteral(str.get(), getter_AddRefs(lit));
  aTree->mDatasource->Assert(v1rsrc, aTree->mValueResource, lit, PR_TRUE);
  aSeq->AppendElement(v1rsrc);

  nsCOMPtr<nsIRDFResource> v2rsrc
    (aTree->GetAssociatedResource("map_variable_var2", mv, nsnull));
  aTree->mDatasource->Assert(v2rsrc, aTree->mTypeResource,
                             aTree->mVariable2Literal, PR_TRUE);
  mv->GetSecondVariableName(str);
  aTree->mRDFService->GetLiteral(str.get(), getter_AddRefs(lit));
  aTree->mDatasource->Assert(v2rsrc, aTree->mValueResource, lit, PR_TRUE);
  aSeq->AppendElement(v2rsrc);
}

void
CellMLToRDFTree::PopulateMapVariablesIntoSequences
(
 cellml_apiIMapVariablesIterator* aMVIt,
 nsIRDFContainer* aMVSeq
)
{
  nsCOMPtr<cellml_apiIMapVariables> mv;
  while (PR_TRUE)
  {
    aMVIt->NextMapVariable(getter_AddRefs(mv));
    if (mv == nsnull)
      break;
    nsCOMPtr<nsIRDFResource> resource
      (GetAssociatedResource("map_variable", mv, &populateMapVariables));

    mDatasource->Assert(resource, mTypeResource, mVariableMappingLiteral,
                        PR_TRUE);
    mDatasource->Assert(resource, mPropResource, mMapVariablesLiteral,
                        PR_TRUE);

    aMVSeq->AppendElement(resource);
  }
}

static void
populateUnit(CellMLToRDFTree* aTree,
             nsISupports* aElement,
             nsIRDFContainer* aSeq)
{
  nsCOMPtr<cellml_apiIUnit> un(do_QueryInterface(aElement));
  nsCOMPtr<nsIRDFResource> prefixResource
    (aTree->GetAssociatedResource("unit_prefix", un, nsnull));
  aTree->mDatasource->Assert(prefixResource, aTree->mTypeResource,
                             aTree->mPrefixLiteral, PR_TRUE);
  PRInt32 value;
  un->GetPrefix(&value);
  nsCOMPtr<nsIRDFInt> ilit;
  aTree->mRDFService->GetIntLiteral(value, getter_AddRefs(ilit));
  aTree->mDatasource->Assert(prefixResource, aTree->mValueResource,
                             ilit, PR_TRUE);
  aSeq->AppendElement(prefixResource);
  
  nsCOMPtr<nsIRDFResource> mupResource
    (aTree->GetAssociatedResource("unit_multiplier", un, nsnull));
  aTree->mDatasource->Assert(mupResource, aTree->mTypeResource,
                             aTree->mMultiplierLiteral, PR_TRUE);
  double dvalue;
  char buf[40];
  un->GetMultiplier(&dvalue);
  PR_snprintf(buf, 40, "%f", dvalue);
  NS_ConvertUTF8toUTF16 wbuf(buf);
  nsCOMPtr<nsIRDFLiteral> lit;
  aTree->mRDFService->GetLiteral(wbuf.get(), getter_AddRefs(lit));
  aTree->mDatasource->Assert(mupResource, aTree->mValueResource, lit,
                             PR_TRUE);
  aSeq->AppendElement(mupResource);
    
  nsCOMPtr<nsIRDFResource> offsResource
    (aTree->GetAssociatedResource("unit_offset", un, nsnull));
  aTree->mDatasource->Assert(offsResource, aTree->mTypeResource,
                             aTree->mOffsetLiteral, PR_TRUE);
  un->GetOffset(&dvalue);
  PR_snprintf(buf, 40, "%f", dvalue);
  NS_ConvertUTF8toUTF16 wbufo(buf);
  aTree->mRDFService->GetLiteral(wbufo.get(), getter_AddRefs(lit));
  aTree->mDatasource->Assert(offsResource, aTree->mValueResource, lit,
                             PR_TRUE);
  aSeq->AppendElement(offsResource);
  
  nsCOMPtr<nsIRDFResource> expResource
    (aTree->GetAssociatedResource("unit_exponent", un, nsnull));
  aTree->mDatasource->Assert(expResource, aTree->mTypeResource,
                             aTree->mExponentLiteral, PR_TRUE);
  un->GetExponent(&dvalue);
  PR_snprintf(buf, 40, "%f", dvalue);
  NS_ConvertUTF8toUTF16 wbufe(buf);
  aTree->mRDFService->GetLiteral(wbufe.get(), getter_AddRefs(lit));
  aTree->mDatasource->Assert(expResource, aTree->mValueResource, lit, PR_TRUE);
  aSeq->AppendElement(expResource);
}

void
CellMLToRDFTree::PopulateUnitIntoSequences
(
 cellml_apiIUnitIterator* aUnIt,
 nsIRDFContainer* aUnitsSeq
)
{
  nsCOMPtr<cellml_apiIUnit> un;
  while (PR_TRUE)
  {
    aUnIt->NextUnit(getter_AddRefs(un));
    if (un == nsnull)
      break;
    nsCOMPtr<nsIRDFResource> resource
      (GetAssociatedResource("unit", un, &populateUnit));

    nsString str;
    un->GetUnits(str);
    nsCOMPtr<nsIRDFLiteral> lit;
    mRDFService->GetLiteral(str.get(), getter_AddRefs(lit));
    mDatasource->Assert(resource, mTypeResource, lit, PR_TRUE);
    mDatasource->Assert(resource, mPropResource, mUnitLiteral, PR_TRUE);

    aUnitsSeq->AppendElement(resource);
  }
}

void
CellMLToRDFTree::PopulateImportComponentsIntoSequences
(
 cellml_apiIImportComponentIterator* aICIt,
 nsIRDFContainer* aICSeq
)
{
  nsCOMPtr<cellml_apiIImportComponent> ic;
  while (PR_TRUE)
  {
    aICIt->NextImportComponent(getter_AddRefs(ic));
    if (ic == nsnull)
      break;
    nsCOMPtr<nsIRDFResource> resource
      (GetAssociatedResource("import_component", ic, nsnull));

    nsString str1, str2;
    ic->GetName(str1);
    ic->GetComponentRef(str2);
    nsCOMPtr<nsIRDFLiteral> lit1, lit2;
    mRDFService->GetLiteral(str1.get(), getter_AddRefs(lit1));
    mRDFService->GetLiteral(str2.get(), getter_AddRefs(lit2));

    mDatasource->Assert(resource, mTypeResource, lit1, PR_TRUE);
    mDatasource->Assert(resource, mValueResource, lit2, PR_TRUE);
    mDatasource->Assert(resource, mPropResource, mComponentsLiteral, PR_TRUE);

    aICSeq->AppendElement(resource);
  }
}

void
CellMLToRDFTree::PopulateImportUnitsIntoSequences
(
 cellml_apiIImportUnitsIterator* aIUIt,
 nsIRDFContainer* aIUSeq
)
{
  nsCOMPtr<cellml_apiIImportUnits> iu;
  while (PR_TRUE)
  {
    aIUIt->NextImportUnits(getter_AddRefs(iu));
    if (iu == nsnull)
      break;
    nsCOMPtr<nsIRDFResource> resource
      (GetAssociatedResource("import_unit", iu, nsnull));

    nsString str1, str2;
    iu->GetName(str1);
    iu->GetUnitsRef(str2);
    nsCOMPtr<nsIRDFLiteral> lit1, lit2;
    mRDFService->GetLiteral(str1.get(), getter_AddRefs(lit1));
    mRDFService->GetLiteral(str2.get(), getter_AddRefs(lit2));

    mDatasource->Assert(resource, mTypeResource, lit1, PR_TRUE);
    mDatasource->Assert(resource, mValueResource, lit2, PR_TRUE);
    mDatasource->Assert(resource, mPropResource, mUnitsLiteral, PR_TRUE);

    aIUSeq->AppendElement(resource);
  }
}

static void
populateVariableRef(CellMLToRDFTree* aTree,
             nsISupports* aElement,
             nsIRDFContainer* aSeq)
{
  nsCOMPtr<cellml_apiIVariableRef> vr(do_QueryInterface(aElement));

  nsCOMPtr<cellml_apiIRoleSet> roles;
  vr->GetRoles(getter_AddRefs(roles));
  nsCOMPtr<cellml_apiIRoleIterator> rolei;
  roles->IterateRoles(getter_AddRefs(rolei));
  aTree->PopulateRolesIntoSequences(rolei, aSeq);
}

void
CellMLToRDFTree::PopulateVariableRefsIntoSequences
(
 cellml_apiIVariableRefIterator* aVRIt,
 nsIRDFContainer* aVRSeq
)
{
  nsCOMPtr<cellml_apiIVariableRef> vr;
  while (PR_TRUE)
  {
    aVRIt->NextVariableRef(getter_AddRefs(vr));
    if (vr == nsnull)
      break;
    nsCOMPtr<nsIRDFResource> resource
      (GetAssociatedResource("variable_ref", vr, &populateVariableRef));

    nsString str;
    vr->GetVariableName(str);
    nsCOMPtr<nsIRDFLiteral> lit;
    mRDFService->GetLiteral(str.get(), getter_AddRefs(lit));
    mDatasource->Assert(resource, mTypeResource, lit, PR_TRUE);
    mDatasource->Assert(resource, mPropResource, mVariableRefLiteral,
                        PR_TRUE);

    aVRSeq->AppendElement(resource);
  }
}

static void
populateRole(CellMLToRDFTree* aTree,
             nsISupports* aElement,
             nsIRDFContainer* aSeq)
{
  nsCOMPtr<cellml_apiIRole> role(do_QueryInterface(aElement));

  nsCOMPtr<nsIRDFResource> dirrsrc
    (aTree->GetAssociatedResource("role_direction", role, nsnull));

  PRUint32 dirtype;
  role->GetDirection(&dirtype);
  nsCOMPtr<nsIRDFLiteral> lit;
  switch (dirtype)
  {
  case cellml_apiIRole::FORWARD:
    lit = aTree->mForwardLiteral;
    break;
  case cellml_apiIRole::REVERSE:
    lit = aTree->mReverseLiteral;
    break;
  case cellml_apiIRole::BOTH:
    lit = aTree->mBothLiteral;
    break;
  }

  aTree->mDatasource->Assert(dirrsrc, aTree->mTypeResource,
                             aTree->mDirectionLiteral, PR_TRUE);
  aTree->mDatasource->Assert(dirrsrc, aTree->mValueResource, lit, PR_TRUE);
  aSeq->AppendElement(dirrsrc);

  nsCOMPtr<nsIRDFResource> stoichrsrc
    (aTree->GetAssociatedResource("role_stoich", role, nsnull));
  aTree->mDatasource->Assert(stoichrsrc, aTree->mTypeResource,
                             aTree->mStoichiometryLiteral, PR_TRUE);
  double dvalue;
  char buf[40];
  role->GetStoichiometry(&dvalue);
  PR_snprintf(buf, 40, "%f", dvalue);
  NS_ConvertUTF8toUTF16 wbuf(buf);
  aTree->mRDFService->GetLiteral(wbuf.get(), getter_AddRefs(lit));
  aTree->mDatasource->Assert(stoichrsrc, aTree->mValueResource, lit, PR_TRUE);
  aSeq->AppendElement(stoichrsrc);

  nsCOMPtr<nsIRDFResource> dvrsrc
    (aTree->GetAssociatedResource("role_deltavariable", role, nsnull));
  aTree->mDatasource->Assert(dvrsrc, aTree->mTypeResource,
                             aTree->mDeltaVariableLiteral, PR_TRUE);
  nsString str;
  role->GetDeltaVariableName(str);
  aTree->mRDFService->GetLiteral(str.get(), getter_AddRefs(lit));
  aTree->mDatasource->Assert(dvrsrc, aTree->mValueResource, lit, PR_TRUE);
  aSeq->AppendElement(dvrsrc);

  nsCOMPtr<cellml_apiIMathContainer> mc(do_QueryInterface(aElement));
  nsCOMPtr<cellml_apiIMathList> ml;
  mc->GetMath(getter_AddRefs(ml));
  nsCOMPtr<cellml_apiIMathMLElementIterator> mmi;
  ml->Iterate(getter_AddRefs(mmi));
  aTree->PopulateMathsIntoSequences(mmi, aSeq);
}

void
CellMLToRDFTree::PopulateRolesIntoSequences
(
 cellml_apiIRoleIterator* aRoleIt,
 nsIRDFContainer* aRoleSeq
)
{
  nsCOMPtr<cellml_apiIRole> role;
  while (PR_TRUE)
  {
    aRoleIt->NextRole(getter_AddRefs(role));
    if (role == nsnull)
      break;
    nsCOMPtr<nsIRDFResource> resource
      (GetAssociatedResource("role", role, &populateRole));

    PRUint32 roletype;
    role->GetVariableRole(&roletype);
    nsCOMPtr<nsIRDFLiteral> lit;
    switch (roletype)
    {
    case cellml_apiIRole::PRODUCT:
      lit = mProductLiteral;
      break;
    case cellml_apiIRole::RATE:
      lit = mRateLiteral;
      break;
    case cellml_apiIRole::CATALYST:
      lit = mCatalystLiteral;
      break;
    case cellml_apiIRole::ACTIVATOR:
      lit = mActivatorLiteral;
      break;
    case cellml_apiIRole::INHIBITOR:
      lit = mInhibitorLiteral;
      break;
    case cellml_apiIRole::MODIFIER:
      lit = mModifierLiteral;
      break;
    case cellml_apiIRole::REACTANT:
    default:
      lit = mReactantLiteral;
      break;
    }

    mDatasource->Assert(resource, mTypeResource, lit, PR_TRUE);
    mDatasource->Assert(resource, mPropResource, mRolesLiteral,
                        PR_TRUE);
    aRoleSeq->AppendElement(resource);
  }
}

static void
populateMaths(CellMLToRDFTree* aTree,
              nsISupports* aElement,
              nsIRDFContainer* aSeq)
{
  nsCOMPtr<mathml_domIMathMLContainer> math(do_QueryInterface(aElement));
  PRUint32 count;
  math->GetNArguments(&count);
  aTree->PopulateEquationsIntoSequences(math, count, aSeq);
}

void
CellMLToRDFTree::PopulateMathsIntoSequences
(
 cellml_apiIMathMLElementIterator* aMEI,
 nsIRDFContainer* aMESeq
)
{
  nsCOMPtr<mathml_domIMathMLElement> me;
  while (PR_TRUE)
  {
    aMEI->Next(getter_AddRefs(me));
    if (me == nsnull)
      break;
    nsCOMPtr<mathml_domIMathMLMathElement> math(do_QueryInterface(me));
    if (math == nsnull)
      continue;
    nsCOMPtr<nsIRDFResource> resource
      (GetAssociatedResource("math", math, &populateMaths));
    mDatasource->Assert(resource, mTypeResource, mMathLiteral, PR_TRUE);
    mDatasource->Assert(resource, mPropResource, mMathLiteral,
                        PR_TRUE);

    // Get the id of the maths...
    nsString cmetaId;
    me->GetAttribute(NS_LITERAL_STRING("id"), cmetaId);
    nsCOMPtr<nsIRDFLiteral> lit;
    mRDFService->GetLiteral(cmetaId.get(), getter_AddRefs(lit));
    mDatasource->Assert(resource, mValueResource, lit, PR_TRUE);

    aMESeq->AppendElement(resource);
  }
}

void
CellMLToRDFTree::PopulateEquationsIntoSequences
(
 mathml_domIMathMLContainer* aContainer,
 PRUint32 aCount,
 nsIRDFContainer* aSeq
)
{
  PRUint32 i;
  nsCOMPtr<mathml_domIMathMLElement> el;
  for (i = 1; i <= aCount; i++)
  {
    aContainer->GetArgument(i, getter_AddRefs(el));
    if (el == nsnull)
      break;
    nsCOMPtr<nsIRDFResource> resource
      (GetAssociatedResource("equation", el, nsnull));
    mDatasource->Assert(resource, mTypeResource, mEquationLiteral, PR_TRUE);
    mDatasource->Assert(resource, mPropResource, mEquationLiteral,
                        PR_TRUE);
    aSeq->AppendElement(resource);
  }
}

PLDHashOperator
CellMLToRDFTree::rememberExpansionCallback
(
 const void* aKey,
 CellMLToRDFTree::ResourceAssociation* aRa,
 void* aInst
)
{
  NS_REINTERPRET_CAST(CellMLToRDFTree*, aInst)->RememberExpansion
    (
     NS_REINTERPRET_CAST(const nsIRDFResource*, aKey),
     aRa
    );
  return PL_DHASH_NEXT;
}

PLDHashOperator
CellMLToRDFTree::serialiseExpansionCallback
(
 const nsACString& aKey,
 int aIgnore,
 void * aInst
)
{
  NS_REINTERPRET_CAST(CellMLToRDFTree*, aInst)->SerialiseExpansion
    (
     aKey
    );
  return PL_DHASH_NEXT;
}

NS_IMETHODIMP
CellMLToRDFTree::RememberExpansions()
{
  mInvertOpenSet.Clear();
  mHash.EnumerateRead(rememberExpansionCallback, this);
  return NS_OK;
}

NS_IMETHODIMP
CellMLToRDFTree::SerialiseExpansions(nsACString& aResult)
{
  mTmpString = NS_LITERAL_CSTRING("");
  mInvertOpenSet.EnumerateRead(serialiseExpansionCallback, this);
  aResult.Assign(mTmpString);

  return NS_OK;
}

void
CellMLToRDFTree::SerialiseExpansion(const nsACString& aKey)
{
  if (!mTmpString.Equals(NS_LITERAL_CSTRING("")))
    mTmpString.Append(NS_LITERAL_CSTRING(","));
  mTmpString.Append(aKey);
}

NS_IMETHODIMP
CellMLToRDFTree::DeserialiseExpansions(const nsACString& aExp)
{
  mInvertOpenSet.Clear();

  const char* data = aExp.BeginReading();
  const char * p, * last = NULL;
  for (p = data - 1; p; p = strchr(p, ','))
  {
    if (last != NULL)
    {
      nsCString assocKey(last, p - last);
      mInvertOpenSet.Put(assocKey, 1);
    }
    last = ++p;
  }
  if (last != NULL)
  {
    nsCString assocKey(last);
    mInvertOpenSet.Put(assocKey, 1);
  }

  return NS_OK;
}

void
CellMLToRDFTree::RememberExpansion(const nsIRDFResource* aRsrc,
                                   ResourceAssociation* aRa)
{
  nsCString assocKey;
  MakeAssociationKey(aRa->mType, aRa->mElement, assocKey);
  PRBool ha = PR_FALSE;
  mPersistSource->HasAssertion(const_cast<nsIRDFResource*>(aRsrc),
                               mOpenResource, mTrueLiteral, true, &ha);
  if (ha != aRa->mDefaultOpen)
    mInvertOpenSet.Put(assocKey, 1);
}

NS_IMETHODIMP
CellMLToRDFTree::GetDatasource
(
 nsIRDFDataSource** aDatasource
)
{
  *aDatasource = mDatasource;
  if (*aDatasource != nsnull)
    NS_ADDREF(*aDatasource);
  return NS_OK;
}

NS_IMETHODIMP
CellMLToRDFTree::GetInformationForResource
(
 nsIRDFResource* aResource,
 nsISupports** aAssociatedElement,
 nsACString& aType
)
{
  ResourceAssociation* ra;
  mHash.Get(aResource, &ra);
  if (ra == nsnull)
    return NS_ERROR_FAILURE;
  aType.Assign(ra->mType);
  *aAssociatedElement = ra->mElement;
  if (*aAssociatedElement != nsnull)
    NS_ADDREF(*aAssociatedElement);
  return NS_OK;
}

NS_IMETHODIMP
CellMLToRDFTree::EnsurePopulated
(
  nsIRDFResource* aTarget
)
{
  ResourceAssociation* ra;

  mHash.Get(aTarget, &ra);
  if (ra == nsnull)
    return NS_ERROR_FAILURE;
  
  if (ra->mPopulate == nsnull)
  {
    return NS_OK;
  }

  nsCOMPtr<nsIRDFNode> ignore;
  ra->mSeq->RemoveElementAt(1, PR_FALSE, getter_AddRefs(ignore));
  rdf_populate_t populate = ra->mPopulate;
  ra->mPopulate = nsnull;

  populate(ra->mCellMLToRDFTree, ra->mElement, ra->mSeq);
  return NS_OK;
}
