const contractID = "@cellml.org/pcenv/content-api;1";
const CID = Components.ID("{5ad45e0e-6953-486b-9262-8d6ecefeeb16}");

pcenvContentAPI.prototype = {
  // Internal
  findAxis: function(axisName)
  {
    if (this.pcenvWindow == null)
      return null;

    if (this.pcenvWindow.axisSet == null)
      return null;

    var i;
    for (i in this.pcenvWindow.axisSet.mAxes)
      if (this.pcenvWindow.axisSet.mAxes[i].name == axisName)
      {
        return this.pcenvWindow.axisSet.mAxes[i];
      }

    return null;
  },

  processColour: function(colour)
  {
    if (colour.match(/^\#[0-9|a-f|A-F]{6}$/) == null)
      return '';

    return colour;
  },

  processDrawType: function(drawType)
  {
    if (drawType == 'lines')
      return 'L';
    else if (drawType == 'squares')
      return 'SQ';
    else if (drawType == 'circles')
      return 'C';
    else if (drawType == 'diamonds')
      return 'DI';
    else if (drawType == 'dots')
      return 'D';
    else if (drawType == 'none')
      return 'I';
    else
      return '';
  },

  // nsISupports
  QueryInterface: function(iid) {
    if (!iid.equals(Components.interfaces.nsISupports) &&
        !iid.equals(Components.interfaces.nsIClassInfo) &&
        !iid.equals(Components.interfaces.IPCEnv) &&
        !iid.equals(Components.interfaces.nsISecurityCheckedComponent))
      throw Components.results.NS_ERROR_NO_INTERFACE;
    return this;
  },

  // nsIClassInfo...
  getInterfaces: function(count) {
    var array = [Components.interfaces.IPCEnv, Components.interfaces.nsIClassInfo,
                 Components.interfaces.nsISecurityCheckedComponent];
    count.value = array.length;
    return array;
  },
  getHelperForLanguage: function(lang) { return null; },
  contractID: contractID,
  classDescription: "The PCEnv content-accessible API",
  classID: CID,
  flags: Components.interfaces.nsIClassInfo.SINGLETON,

  // nsISecurityCheckedComponent
  canCreateWrapper: function(iid) {
    return "allAccess";
  },

  canCallMethod: function(iid, methodName) {
    if (methodName == "setPCEnvWindow" ||
        !iid.equals(Components.interfaces.IPCEnv))
      return "noAccess";
    else
      return "allAccess";
  },

  canGetProperty: function(iid, propertyName) {
    return "NoAccess";
  },

  canSetProperty: function(iid, propertyName) {
    return "NoAccess";
  },

  // IPCEnv...

  status: function(message) {
    this.pcenvWindow.document.getElementById("statusbar").label =
      "Script Message: " + message;
  },

  hideTraces: function(axisName) {
    var axis = this.findAxis(axisName);
    if (axis == null)
      return;

    var locked = axis.locked;
    if (locked)
    {
      axis.lockedBy.key.setAllDrawTypes('I');
      return;
    }

    var i;
    for (i in axis.lines)
    {
      axis.lines[i].drawType = 'I';
    }
  },

  selectTrace: function(axisName, xname, yname, colour, drawType) {
    var axis = this.findAxis(axisName);
    if (axis == null)
    {
      // dump("Axis not found\n");
      return;
    }

    drawType = this.processDrawType(drawType);
    colour = this.processColour(colour);

    // dump("drawType = " + drawType + " and colour = " + colour + "\n");

    var i;
    for (i in axis.lines)
    {
      var l = axis.lines[i];

      if (
          l.xvar == null ||
          l.xcomp == null ||
          l.yvar == null ||
          l.ycomp == null
         )
      {
        // dump("Skipping line because it has null variables or components.\n");
        continue;
      }
      
      if (l.xcomp + '/' + l.xvar != xname ||
          l.ycomp + '/' + l.yvar != yname)
      {
        // dump("Skipping line because mismatch on " +
        //      l.xcomp + '/' + l.xvar + " vs " + xname +
        //      " or " + l.ycomp + '/' + l.yvar + ' vs ' + yname + "\n");
        continue;
      }

      if (!axis.locked)
      {
        if (colour != "")
        {
          l.setColour(colour);

          // Now we have to do this unfortunate hack to refresh the CSS colour...
          var p = "prop";
          for (i = 0; i < 10; i++)
            p = p + "0123456789ABCDEF"[parseInt(Math.random() * 16)];
          l.properties = p;

          var i;
          for (i = 0; i < this.pcenvWindow.document.styleSheets.length; i++)
          {
            if (this.pcenvWindow.document.styleSheets[i].href ==
                "chrome://pcenv/skin/graphkey.css")
              break;
          }
          var sheet = this.pcenvWindow.document.styleSheets[i];
          var r = "treechildren::-moz-tree-cell(" + p + ") { background-color: " +
            colour + "; }";
          sheet.insertRule(r, sheet.cssRules.length);
        }

        if (drawType != "")
          l.setDrawType(drawType);

        // dump("Trace was fixed while not in use.\n");

        return;
      }

      // dump("Trace was in use, about to update it live.\n");

      // The axis is displayed on the screen live. Set the line type...
      axis.lockedBy.key.setDrawInformationByLine(l, colour, drawType);

      return;
    }
    // dump("Trace not found\n");
  },

  setPCEnvWindow: function(pcenvWindow) {
    this.pcenvWindow = pcenvWindow;
  }
};

function pcenvContentAPI()
{
}

var PCEnvContentAPIModule = {
  registered: false,

  registerSelf: function(compMgr, fileSpec, location, type)
  {
    if (this.registered)
      return;

    var compReg = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    compReg.registerFactoryLocation(CID, "PCEnv Content API", contractID,
                                    fileSpec, location, type);
    var catman = Components.classes['@mozilla.org/categorymanager;1']
                   .getService(Components.interfaces.nsICategoryManager);
    catman.addCategoryEntry("JavaScript global property", "pcenv",
                            contractID, true, true);
  },

  unregisterSelf: function(compMgr, location, type)
  {
    var catman = Components.classes['@mozilla.org/categorymanager;1']
                   .getService(Components.interfaces.nsICategoryManager);
    catman.deleteCategoryEntry("JavaScript global property", "pcenv",
                               true);
  },

  getClassObject: function(compMgr, cid, iid) {
    if (cid.equals(CID))
      return this;
    throw Components.results.NS_ERROR_NO_INTERFACE;
  },

  createInstance: function mod_CI(outer, iid) {
    if (outer != null)
      throw Components.results.NS_ERROR_NO_AGGREGATION;
    // Enforce singleton...
    if (this.pcenvContentAPI == null)
    {
      this.pcenvContentAPI = new pcenvContentAPI();
    }

    return this.pcenvContentAPI.QueryInterface(iid);
  }
};

function NSGetModule(cm, file)
{
  return PCEnvContentAPIModule;
}
