function getModelMetadata(model)
{
  try
  {
    var sd = model.getRDFRepresentation("http://www.cellml.org/RDFXML/string").
      QueryInterface(Components.interfaces.
                     cellml_apiIRDFXMLStringRepresentation).serialisedData;
    
    // Put it into the Mozilla RDF implementation...
    var p = Components.classes["@mozilla.org/rdf/xml-parser;1"].
      createInstance(Components.interfaces.nsIRDFXMLParser);
    var mds = Components.classes["@mozilla.org/rdf/datasource;1?name=in-memory-datasource"].
      createInstance(Components.interfaces.nsIRDFDataSource);
    var modelURI = model.base_uri.asText;
    var uri = Components.classes["@mozilla.org/network/standard-url;1"].
      createInstance(Components.interfaces.nsIURI);
    uri.spec = modelURI;
    p.parseString(mds, uri, sd);
    return mds;
  }
  catch (e)
  {
    alert("There was a problem getting the metadata: " + e);
    return null;
  }
}

function setModelMetadata(model, datasource)
{
  // Firstly serialise the RDF into a mini-document...
  var xs = Components.classes["@mozilla.org/rdf/xml-serializer;1"].
    createInstance(Components.interfaces.nsIRDFXMLSerializer);
  xs.QueryInterface(Components.interfaces.nsIRDFXMLSource).init(datasource);
  var outputStream = {
    data: "",
    close : function(){},
    flush : function(){},
    write : function (buffer,count){
      this.data += buffer;
      return count;
    },
    writeFrom : function (stream,count){},
    isNonBlocking: false
  }
  xs.Serialize(outputStream);

  model.getRDFRepresentation("http://www.cellml.org/RDFXML/string").
    QueryInterface(Components.interfaces.
                   cellml_apiIRDFXMLStringRepresentation).serialisedData =
        outputStream.data;
}

// Changes the base URI and updates metadata...
function changeBaseURI(model, newBase)
{
  newBase = newBase.QueryInterface(Components.interfaces.nsIURL);
  var ios = Components.classes["@mozilla.org/network/io-service;1"].
    getService(Components.interfaces.nsIIOService);
  try
  {
    var oldBase = ios.newURI(model.base_uri.asText, "UTF8", null);
  } catch (e)
  {
    // Give up if no valid xml:base present.
    return;
  }
  oldBase = oldBase.QueryInterface(Components.interfaces.nsIURL);

  var ds = getModelMetadata(model);
  var dsOut = Components.classes["@mozilla.org/rdf/datasource;1?name=in-memory-datasource"].
      createInstance(Components.interfaces.nsIRDFDataSource);

  var rtv =
    {
      QueryInterface: function(IID)
      {
        if (IID == Components.interfaces.nsISupports ||
            IID == Components.interfaces.rdfITripleVisitor)
          return this;
        throw Components.results.NS_ERROR_NO_INTERFACE;
      },
      
      considerTransform: function(input)
      {
        var r;
        try
        {
          r = input.QueryInterface(Components.interfaces.nsIRDFResource);
        }
        catch (e)
        {
          return input;
        }
        if (r == null)
          return input;

        var rURI = ios.newURI(r.ValueUTF8, "UTF8", null);
        try
        {
          rURI = rURI.QueryInterface(Components.interfaces.nsIURL);
        }
        catch (e)
        {
          return input;
        }

        if (rURI.prePath != oldBase.prePath ||
            rURI.filePath != oldBase.filePath)
          return input;

        this.dirty = true;
        rURI.prePath = newBase.prePath;
        rURI.filePath = newBase.filePath;

        return window.rdfservice.GetResource(rURI.spec);
      },

      visit: function(subject, predicate, object, truth)
      {
        if (!truth)
          return;
        
        subject = this.considerTransform(subject);
        predicate = this.considerTransform(predicate);
        object = this.considerTransform(object);

        rtv.ds.Assert(subject, predicate, object, truth);
      },

      ds: dsOut,
      dirty: false
    };

  var rds = ds.QueryInterface(Components.interfaces.rdfIDataSource);
  rds.visitAllTriples(rtv);

  if (rtv.dirty)
    setModelMetadata(model, rtv.ds);

  model.base_uri.asText = newBase.spec;
}
