#!/usr/bin/python
import sys, os, shutil, re

Paths={}

def EnterContext(ctx):
    contextStack.append(ctx)
    
class ToplevelProcessor:
    def __init__(self):
        self.level = 0
    def gotDirective(self, line):
        if line == "Copy:":
            EnterContext(CopyProcessor())
        elif line[0:8] == "Execute ":
            command = line[8:]
            command = re.sub('\$[A-Z|a-z|0-9]+',\
                             lambda x: Paths[x.group(0)[1:]], command)
            command = re.sub('[a-z|A-Z]+\/\.\./([a-z|A-Z]+)',\
                             '\\1', command)
            command = re.sub('[a-z|A-Z]+\/\.\.', '', command)
            os.system(command)
        else:
            print "Got unknown top-level directive: \"" + line + "\""
    def endDirective(self):
        pass
    
class CopyProcessor:
    def __init__(self):
        self.level = -1
    def gotDirective(self, line):
        if line[0:5] == 'From ' and line[-1] == ':':
            srcproject=line[5:-1]
            EnterContext(CopyFromProcessor(srcproject))
        else:
            print "Got unsupported copy directive: " + line
    def endDirective(self):
        pass

class CopyFromProcessor:
    def __init__(self, srcproject, destdir=''):
        self.level = -1
        self.origsrcproject = srcproject
        parts = srcproject.split('/')
        self.destdir = Paths['stagingDir']
        if destdir != '':
            self.destdir += '/' + destdir
        try:
            os.makedirs(self.destdir)
        except OSError:
            pass
        self.srcproject = Paths[parts[0]]
        for p in parts[1:]:
            self.srcproject += '/'
            self.srcproject += p
    def gotDirective(self, line):
        if line[0:3] == "To " and line[-1] == ":":
            destdir=line[3:-1]
            EnterContext(CopyFromProcessor(self.origsrcproject, destdir))
        else:
            src = self.srcproject + "/" + line
            dest = self.destdir + "/" + line
            shutil.copyfile(src, dest)
            shutil.copymode(src, dest)
    def endDirective(self):
        pass

contextStack = [ToplevelProcessor()]

if len(sys.argv) < 2:
    print "Usage: FinalStageMaker platform.spec"
    sys.exit(-1)

for option in sys.argv[2:]:
    optiona=option.split('=')
    if len(optiona) < 2:
        print 'Warning: Expected = in argument ' + option
        continue
    Paths[optiona[0]] = optiona[1]
if not Paths.has_key('stagingDir'):
    Paths['stagingDir'] = '/tmp/pcenvStage'

if os.path.isdir(Paths['stagingDir']):
    for root, dirs, files in os.walk(Paths['stagingDir'], topdown=False):
        for name in files:
            os.remove(os.path.join(root, name))
        for name in dirs:
            os.rmdir(os.path.join(root, name))
else:
    os.mkdir(Paths['stagingDir'])

input = file(sys.argv[1], "r")
lineno = 0
for line in input:
    lineno = lineno + 1
    line=line.rstrip()
    sline=line.lstrip()
    indent = len(line)-len(sline)
    while contextStack[-1].level > indent:
        contextStack.pop().endDirective()

    if contextStack[-1].level == -1:
        contextStack[-1].level = indent
    
    if indent != contextStack[-1].level:
        print "Warning: Unexpected indent at line number %u" % lineno
    if sline != "pass" and sline[0] != '#':
        contextStack[-1].gotDirective(sline)
    
while len(contextStack):
    contextStack.pop().endDirective()
